from ._deformation import DeformationSpace
from ._dilatational import DilatationalSpace
from ._distortional import DistortionalSpace

__all__ = ["DistortionalSpace", "DilatationalSpace", "DeformationSpace"]
